#include "asterx1_gps.h"
#include "asterx1exceptions.h"
#include "gps_types.h"
#include "ctime.h"

/* Look up table for fast computation of the CCITT 16-bit CRC. */
static const unsigned short CRC_16CCIT_LookUp[256] = {
  0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
  0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
  0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
  0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
  0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
  0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
  0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
  0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
  0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
  0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
  0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
  0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
  0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
  0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
  0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
  0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
  0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
  0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
  0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
  0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
  0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
  0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
  0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
  0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
  0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
  0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
  0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
  0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
  0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
  0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
  0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
  0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};

CasteRx1::CasteRx1(const std::string &name,const std::string &hwPortName)
{
  this->serialPort=NULL;
  this->serial_device=hwPortName;
  this->gps_running=false;
  /* initialize events */
  this->event_server=CEventServer::instance();
  this->finish_thread_event_id=name+"_finish_thread_event";
  this->event_server->create_event(this->finish_thread_event_id);
  this->new_reply_event_id=name+"_new_reply_event";
  this->event_server->create_event(this->new_reply_event_id);
  this->new_meas_block_event_id=name+"_new_meas_block";
  this->event_server->create_event(this->new_meas_block_event_id);
  this->new_gps_nav_block_event_id=name+"_new_gps_nav_block";
  this->event_server->create_event(this->new_gps_nav_block_event_id);
  this->new_gps_alm_block_event_id=name+"_new_gps_alm_block";
  this->event_server->create_event(this->new_gps_alm_block_event_id);
  this->new_gps_ion_block_event_id=name+"_new_gps_ion_block";
  this->event_server->create_event(this->new_gps_ion_block_event_id);
  this->new_gps_utc_block_event_id=name+"_new_gps_utc_block";
  this->event_server->create_event(this->new_gps_utc_block_event_id);
  this->new_gps_raw_block_event_id=name+"_new_gps_raw_block";
  this->event_server->create_event(this->new_gps_raw_block_event_id);
  this->new_pvt_block_event_id=name+"_new_pvt_block";
  this->event_server->create_event(this->new_pvt_block_event_id);
  /* initialize threads */
  this->thread_server=CThreadServer::instance();
  this->gps_process_data_thread_id=name+"_gps_process_data_thread";
  this->thread_server->create_thread(this->gps_process_data_thread_id);
  this->thread_server->attach_thread(this->gps_process_data_thread_id,this->gps_process_data_thread,this);
  this->sat_channel.resize(256,-1);
}

CasteRx1::~CasteRx1()
{
  // stop data acquiaition
  this->stopAcquisition();
  // close the gps module
  this->closeDevice();
  // delete threads and events
  this->thread_server->detach_thread(gps_process_data_thread_id);
  this->thread_server->delete_thread(gps_process_data_thread_id);
  this->gps_process_data_thread_id="";
  this->event_server->delete_event(this->finish_thread_event_id);
  this->finish_thread_event_id="";
  this->event_server->delete_event(this->new_reply_event_id);
  this->new_reply_event_id="";
  this->event_server->delete_event(this->new_meas_block_event_id);
  this->new_meas_block_event_id="";
  this->event_server->delete_event(this->new_gps_nav_block_event_id);
  this->new_gps_nav_block_event_id="";
  this->event_server->delete_event(this->new_gps_alm_block_event_id);
  this->new_gps_alm_block_event_id="";
  this->event_server->delete_event(this->new_gps_ion_block_event_id);
  this->new_gps_ion_block_event_id="";
  this->event_server->delete_event(this->new_gps_utc_block_event_id);
  this->new_gps_utc_block_event_id="";
  this->event_server->delete_event(this->new_gps_raw_block_event_id);
  this->new_gps_raw_block_event_id="";
  this->event_server->delete_event(this->new_pvt_block_event_id);
  this->new_pvt_block_event_id="";
  if(this->serialPort!=NULL)
  {
    delete this->serialPort;
    this->serialPort=NULL;
  }
}

void CasteRx1::openDevice()
{
  TRS232_config serialConfig;

  try
  {
    this->serialPort = new CRS232(this->serial_device); //creates a new object to manage serial connection
    this->serialPort->open((void*)&this->serial_device); 
    serialConfig.baud = 115200;
    serialConfig.num_bits = 8;
    serialConfig.parity = none;
    serialConfig.stop_bits = 1;
    this->serialPort->config(&serialConfig);
    // start the GPS processing data thread
    this->thread_server->start_thread(this->gps_process_data_thread_id);
    // configure the data streams
    this->configureStreams();
  }catch(CCommException &e){
    if(this->serialPort!=NULL)
    {
      delete this->serialPort;
      this->serialPort=NULL;
    }
    throw CasteRx1Exception(_HERE_,"Impossible to open the communications device");
  }
}

void CasteRx1::closeDevice()
{
  // stop the internal acquisition thread if necessary
  this->stopAcquisition();
  // close the serial port
  if(this->thread_server->get_thread_state(this->gps_process_data_thread_id)==active)
  {
    this->event_server->set_event(this->finish_thread_event_id);
    this->thread_server->end_thread(this->gps_process_data_thread_id);
  }
  if(this->serialPort!=NULL)
    this->serialPort->close(); 
}

void CasteRx1::startAcquisition()
{
  if(this->thread_server->get_thread_state(this->gps_process_data_thread_id)==active)
  {
    if(!this->gps_running)
    {
      this->sendCommand("setSBFOutput,Stream1,USB1,Group1,OnChange\n");
      this->sendCommand("setSBFOutput,Stream2,USB1,Group2,OnChange\n");
      this->sendCommand("setSBFOutput,Stream3,USB1,Group3,OnChange\n");
      this->sendCommand("setSBFOutput,Stream4,USB1,Group4,OnChange\n");
      this->gps_running=true;
    }
  }
}

void CasteRx1::stopAcquisition()
{
  if(this->gps_running)
  {
    this->sendCommand("setSBFOutput,Stream1,USB1,none\n");
    this->sendCommand("setSBFOutput,Stream2,USB1,none\n");
    this->sendCommand("setSBFOutput,Stream3,USB1,none\n");
    this->sendCommand("setSBFOutput,Stream4,USB1,none\n");
    this->gps_running=false;
  }
}

std::string CasteRx1::get_new_meas_data_event_id(void)
{
  return this->new_meas_block_event_id;
}

void CasteRx1::get_meas_data(TMeasEpoch &meas_epoch, TMeasExtra &meas_extra)
{
  this->meas_access.enter();
  meas_epoch=this->current_meas_epoch;
  meas_extra=this->current_meas_extra;
  if(this->event_server->event_is_set(this->new_meas_block_event_id))
    this->event_server->reset_event(this->new_meas_block_event_id);
  this->meas_access.exit();
}

std::string CasteRx1::get_new_gps_nav_data_event_id(void)
{
  return this->new_gps_nav_block_event_id;
}

void CasteRx1::get_gps_nav_data(TGPSNav &gps_nav)
{
  this->nav_access.enter();
  gps_nav=this->current_gps_nav;
  if(this->event_server->event_is_set(this->new_gps_nav_block_event_id))
    this->event_server->reset_event(this->new_gps_nav_block_event_id);
  this->nav_access.exit(); 
}

std::string CasteRx1::get_new_gps_alm_data_event_id(void)
{
  return this->new_gps_alm_block_event_id;
}

void CasteRx1::get_gps_alm_data(TGPSAlm &gps_alm)
{
  this->nav_access.enter();
  gps_alm=this->current_gps_alm;
  if(this->event_server->event_is_set(this->new_gps_alm_block_event_id))
    this->event_server->reset_event(this->new_gps_alm_block_event_id);
  this->nav_access.exit(); 
}

std::string CasteRx1::get_new_gps_ion_data_event_id(void)
{
  return this->new_gps_ion_block_event_id;
}

void CasteRx1::get_gps_ion_data(TGPSIon &gps_ion)
{
  this->nav_access.enter();
  gps_ion=this->current_gps_ion;
  if(this->event_server->event_is_set(this->new_gps_ion_block_event_id))
    this->event_server->reset_event(this->new_gps_ion_block_event_id);
  this->nav_access.exit(); 
}

std::string CasteRx1::get_new_gps_utc_data_event_id(void)
{
  return this->new_gps_utc_block_event_id;
}

void CasteRx1::get_gps_utc_data(TGPSUtc &gps_utc)
{
  this->nav_access.enter();
  gps_utc=this->current_gps_utc;
  if(this->event_server->event_is_set(this->new_gps_utc_block_event_id))
    this->event_server->reset_event(this->new_gps_utc_block_event_id);
  this->nav_access.exit(); 
}

std::string CasteRx1::get_new_gps_raw_data_event_id(void)
{
  return this->new_gps_raw_block_event_id;
}

void CasteRx1::get_gps_raw_data(TGPSFrame &gps_frame)
{
  this->nav_access.enter();
  if(!this->gps_raw_frames.empty())
  {
    gps_frame=this->gps_raw_frames.front();
    this->gps_raw_frames.pop();
    if(this->gps_raw_frames.empty())
    {
      if(this->event_server->event_is_set(this->new_gps_raw_block_event_id))
        this->event_server->reset_event(this->new_gps_raw_block_event_id);
    }
  }
  this->nav_access.exit(); 
}

std::string CasteRx1::get_new_pvt_data_event_id(void)
{
  return this->new_pvt_block_event_id;
}

void CasteRx1::get_pvt_data(TPVTCartesian &pvt_cart,TPVTGeodetic &pvt_geo)
{
  this->pvt_access.enter();
  pvt_cart=this->current_pvt_cartesian;
  pvt_geo=this->current_pvt_geodetic;
  if(this->event_server->event_is_set(this->new_pvt_block_event_id))
    this->event_server->reset_event(this->new_pvt_block_event_id);
  this->pvt_access.exit(); 
}

void CasteRx1::get_pvt_cov_data(TPVTCov &pvt_pos_cov_cart,TPVTCov &pvt_pos_cov_geo,TPVTCov &pvt_vel_cov_cart,TPVTCov &pvt_vel_cov_geo)
{
  this->pvt_access.enter();
  pvt_pos_cov_cart=this->current_pvt_pos_cov_cartesian;
  pvt_pos_cov_geo=this->current_pvt_pos_cov_geodetic;
  pvt_vel_cov_cart=this->current_pvt_vel_cov_cartesian;
  pvt_vel_cov_geo=this->current_pvt_vel_cov_geodetic;
  if(this->event_server->event_is_set(this->new_pvt_block_event_id))
    this->event_server->reset_event(this->new_pvt_block_event_id);
  this->pvt_access.exit(); 
}

void CasteRx1::get_pvt_dop_data(TPVTDOP &pvt_dop)
{
  this->pvt_access.enter();
  pvt_dop=this->current_pvt_dop;
  if(this->event_server->event_is_set(this->new_pvt_block_event_id))
    this->event_server->reset_event(this->new_pvt_block_event_id);
  this->pvt_access.exit(); 
}

/******************************* PROTECTED MEMBERS ************************************************/

void *CasteRx1::gps_process_data_thread(void *params)
{
  unsigned int event_id,num,i,length=0;
  CasteRx1 *gps=(CasteRx1 *)params;
  static gps_state_t state=header1;
  std::list<std::string> events;
  unsigned char *data,*sbf_data;
  std::string reply;
  bool end=false;
  
  events.push_back(gps->finish_thread_event_id);
  events.push_back(gps->serialPort->get_rx_event_id());
  while(!end)
  {
    event_id=gps->event_server->wait_first(events);
    if(event_id==0)// finish the thread
      end=true;
    else if(event_id==1)// new data available
    {
      num=gps->serialPort->get_num_data(); 
      if(data!=NULL)
        delete[] data;
      data=new unsigned char[num];
      gps->serialPort->read(data,num);
      for(i=0;i<num;i++)
      {
        switch(state)
        {
          case header1: if(data[i]=='$')
                          state=header2;
                        else
                          state=header1;
                        break;
          case header2: if(data[i]=='R')// answer to a command
                          state=get_cmd_reply;
                        else if(data[i]=='@')// gps data
                        {
                          state=get_block_header;
                          length=2;
                          gps->current_header.sync[0]='$';
                          gps->current_header.sync[1]='@';
                        }
                        break;
          case get_cmd_reply: if(data[i]=='\r')
                              {
                                // signal that an answer is available
                                gps->reply_queue.push(reply);
                                if(!gps->event_server->event_is_set(gps->new_reply_event_id))
                                  gps->event_server->set_event(gps->new_reply_event_id);
                                reply="";
                                state=header1;
                              }
                              else// add the data to the answer string
                              {
                                reply.push_back(data[i]);
                                state=get_cmd_reply;
                              }
                              break;
          case get_block_header: ((unsigned char *)&gps->current_header)[length]=data[i];
                                 length++;
                                 if(length==sizeof(TGPSBlockHeader))
                                 {
                                   if((gps->current_header.length%4)!=0)// length must be a multiple of 4
                                     state=header1;
                                   else
                                   {
                                     if(sbf_data!=NULL)
                                       delete sbf_data;
                                     sbf_data=new unsigned char[gps->current_header.length];
                                     state=get_sbf_data;
                                   }
                                 }
                                 break;
          case get_sbf_data: sbf_data[length-sizeof(TGPSBlockHeader)]=data[i];
                             length++;
                             if(length==gps->current_header.length)
                             {
                               /* compute CRC and check it */
                               if(gps->compute_crc(sbf_data,length-8)==gps->current_header.crc)// invalid Block CRC
                                 gps->process_block(sbf_data,length-8);
                               else
                                 std::cout << "CRC error" << std::endl;
                               delete[] sbf_data;
                               sbf_data=0;
                               state=header1;
                             }
                             break;
        }
      }
      delete[] data;
      data=NULL;
    }
  }
  pthread_exit(NULL);
}

void CasteRx1::process_block(unsigned char *data,unsigned short int length)
{
  switch(this->current_header.id.id)
  {
    case PVTCartesian_ID: //std::cout << "PVTCartesian_ID" << std::endl;
                          this->process_pvt_cartesian(data,length);
			  break;
    case PVTGeodetic_ID: //std::cout << "PVTGeodetic_ID" << std::endl;
                         this->process_pvt_geodetic(data,length);
			 break;
    case PosCovCartesian_ID: //std::cout << "PosCovCartesian_ID" << std::endl;
                             this->process_pos_cov_cartesian(data,length);
			     break;
    case PosCovGeodetic_ID: //std::cout << "PosCovGeodetic_ID" << std::endl;
                            this->process_pos_cov_geodetic(data,length);
			    break;
    case VelCovCartesian_ID: //std::cout << "VelCovCartesian_ID" << std::endl;
                             this->process_vel_cov_cartesian(data,length);
			     break;
    case VelCovGeodetic_ID: //std::cout << "VelCovGeodetic_ID" << std::endl;
                            this->process_vel_cov_geodetic(data,length);
			    break;
    case DOP_ID: //std::cout << "DOP_ID" << std::endl;
                 this->process_dop(data,length);
		 break;
    case EndOfPVT_ID: //std::cout << "EndofPVT_ID" << std::endl;
                      this->process_end_of_pvt(data,length);
		      break;
    case MeasEpoch_ID: //std::cout << "MeasEpoch_ID" << std::endl;
                       this->process_meas_epoch(data,length);
		       break;
    case MeasExtra_ID: //std::cout << "MeasExtra_ID" << std::endl;
                       this->process_meas_extra(data,length);
		       break;
    case EndOfMeas_ID: //std::cout << "EndofMeas_ID" << std::endl;
                       this->process_end_of_meas(data,length);
		       break;
    case GPSNav_ID: //std::cout << "GPSNav_ID" << std::endl;
                    this->process_gps_nav(data,length);
		    break;
    case GPSAlm_ID: //std::cout << "GPSAlm_ID" << std::endl;
                    this->process_gps_alm(data,length);
		    break;
    case GPSIon_ID: //std::cout << "GPSIon_ID" << std::endl;
                    this->process_gps_ion(data,length);
		    break;
    case GPSUtc_ID: //std::cout << "GPSUtc_ID" << std::endl;
                    this->process_gps_utc(data,length);
		    break;
    case GPSRawCA_ID: //std::cout << "GPSRawCA_ID" << std::endl;
                      this->process_gps_raw_ca(data,length);
		      break;
    case ReceiverStatus_ID: //std::cout << "ReceiverStatus_ID" << std::endl;
                            this->process_receiver_status(data,length);
			    break;
    case ChannelStatus_ID: //std::cout << "ChannelStatus_ID" << std::endl;
			   this->process_channel_status(data,length);
			   break;
    case SatVisibility_ID: //std::cout << "SatVisibility_ID" << std::endl;
                           this->process_sat_visibility(data,length);
			   break;
    default: std::cout << "Unknown block ID: " << this->current_header.id.id << std::endl;
	     break;
  }
}

void CasteRx1::process_channel_status(unsigned char *data,unsigned short int length)
{
  TGPSChannelStatus gps_channel_status;
  TGPSChannelSatInfo gps_sat_info;
  TGPSChannelStateInfo gps_state_info;
  unsigned short int index=0,i,j;
  TChannelStatus channel_status;
  TChannelSatInfo sat_info;
  TChannelStateInfo state_info;
  
  memcpy((unsigned char *)&gps_channel_status,data,sizeof(TGPSChannelStatus));
  index+=sizeof(TGPSChannelStatus);
  channel_status.gps_timestamp.wnc=gps_channel_status.timestamp.wnc;
  channel_status.gps_timestamp.tow=gps_channel_status.timestamp.tow;
  channel_status.local_timestamp=this->local_time.getTimeInSeconds();
  channel_status.info.clear();
  for(i=0;i<gps_channel_status.num_ch;i++)
  {
    memcpy((unsigned char *)&gps_sat_info,&data[index],sizeof(TGPSChannelSatInfo));
    index+=gps_channel_status.length_ch_sat_info; 
    sat_info.sat_id=gps_sat_info.sat_id;
    sat_info.azimuth=gps_sat_info.azimuth_rise_set.azimuth;
    sat_info.rise_set=(rise_set_t)gps_sat_info.azimuth_rise_set.rise_set;
    sat_info.health=(health_t)gps_sat_info.health;
    sat_info.elevation=gps_sat_info.elevation;
    sat_info.rx_channel=gps_sat_info.rx_channel;
    this->sat_channel[sat_info.sat_id]=gps_sat_info.rx_channel;
    sat_info.state.clear();
    for(j=0;j<gps_sat_info.num_state_info;j++)
    { 
      memcpy((unsigned char *)&gps_state_info,&data[index],sizeof(TGPSChannelStateInfo));
      index+=gps_channel_status.length_ch_state_info;
      state_info.antenna_id=(antenna_t)gps_state_info.antenna;
      state_info.track_status=(track_status_t)gps_state_info.tracking_status;
      state_info.pvt_status=(pvt_status_t)gps_state_info.pvt_status;
      sat_info.state.push_back(state_info);
    }
  }
}

void CasteRx1::process_receiver_status(unsigned char *data,unsigned short int length)
{
  TGPSReceiverStatus gps_receiver_status;
  unsigned short int index=0,i;
  TGPSAGCData gps_agc_data;
  TAGCData agc_data;

  memcpy((unsigned char *)&gps_receiver_status,data,sizeof(TGPSReceiverStatus));
  index+=sizeof(TGPSReceiverStatus);
  this->receiver_status.gps_timestamp.wnc=gps_receiver_status.timestamp.wnc;
  this->receiver_status.gps_timestamp.tow=gps_receiver_status.timestamp.tow;
  this->receiver_status.local_timestamp=this->local_time.getTimeInSeconds();
  this->receiver_status.cpu_load=gps_receiver_status.cpu_load;
  this->receiver_status.up_time=gps_receiver_status.up_time;
  // state flags
  this->receiver_status.state=*((int*)&gps_receiver_status.rx_state);
  this->receiver_status.error=*((int*)&gps_receiver_status.rx_error);
  this->receiver_status.agc.clear();
  for(i=0;i<gps_receiver_status.num_agc_states;i++)
  {
    memcpy((unsigned char *)&gps_agc_data,&data[index],sizeof(TGPSAGCData));
    index+=gps_receiver_status.length_agc_states;
    agc_data.frontend_id=(frontend_t)gps_agc_data.frontend_id.code;
    agc_data.antenna_id=(antenna_t)gps_agc_data.frontend_id.antenna_id;
    agc_data.gain=gps_agc_data.gain;
    agc_data.sample_var=gps_agc_data.sample_var;
    agc_data.blanking_stat=gps_agc_data.blanking_stat;
    this->receiver_status.agc.push_back(agc_data);
  }
}

void CasteRx1::process_sat_visibility(unsigned char *data,unsigned short int length)
{
  TGPSSatVisibility gps_sat_visibility;
  unsigned short int index=0,i;
  TGPSSatInfo gps_sat_info;

  memcpy((unsigned char *)&gps_sat_visibility,data,sizeof(TGPSSatVisibility));
  index+=sizeof(TGPSSatVisibility);
  for(i=0;i<gps_sat_visibility.num_sat;i++)
  {
    memcpy((unsigned char *)&gps_sat_info,&data[index],sizeof(TGPSSatInfo));
    index+=gps_sat_visibility.length_sat;
    this->sat_info.gps_timestamp.wnc=gps_sat_visibility.timestamp.wnc;
    this->sat_info.gps_timestamp.tow=gps_sat_visibility.timestamp.tow;
    this->sat_info.local_timestamp=this->local_time.getTimeInSeconds();
    this->sat_info.sat_id=gps_sat_info.sat_id;
    this->sat_info.azimuth=gps_sat_info.azimuth*0.01;
    this->sat_info.elevation=gps_sat_info.elevation*0.01;
    this->sat_info.rise_set=(rise_set_t)gps_sat_info.rise_set;
    this->sat_info.sat_info=(sat_info_t)gps_sat_info.sat_info;
  }
}

void CasteRx1::process_gps_nav(unsigned char *data,unsigned short int length)
{
  TGPSGPSNav gps_gps_nav;

  this->nav_access.enter();
  memcpy((unsigned char *)&gps_gps_nav,data,sizeof(TGPSGPSNav));
  this->current_gps_nav.gps_timestamp.wnc=gps_gps_nav.timestamp.wnc;
  this->current_gps_nav.gps_timestamp.tow=gps_gps_nav.timestamp.tow;
  this->current_gps_nav.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_gps_nav.sat_id=gps_gps_nav.prn;
  this->current_gps_nav.wn=gps_gps_nav.wn;
  this->current_gps_nav.ca_or_pon_l2=gps_gps_nav.ca_or_pon_l2;
  this->current_gps_nav.ura=gps_gps_nav.ura;
  this->current_gps_nav.health=gps_gps_nav.health;
  this->current_gps_nav.l2_data_flag=gps_gps_nav.l2_data_flag;
  this->current_gps_nav.iodc=gps_gps_nav.iodc;
  this->current_gps_nav.iode2=gps_gps_nav.iode2;
  this->current_gps_nav.iode3=gps_gps_nav.iode3;
  this->current_gps_nav.fit_int_flag=gps_gps_nav.fit_int_flag;
  this->current_gps_nav.t_gd=gps_gps_nav.t_gd;
  this->current_gps_nav.t_oc=gps_gps_nav.t_oc;
  this->current_gps_nav.a_f2=gps_gps_nav.a_f2;
  this->current_gps_nav.a_f1=gps_gps_nav.a_f1;
  this->current_gps_nav.a_f0=gps_gps_nav.a_f0;
  this->current_gps_nav.c_rs=gps_gps_nav.c_rs;
  this->current_gps_nav.delta_n=gps_gps_nav.delta_n;
  this->current_gps_nav.m_0=gps_gps_nav.m_0;
  this->current_gps_nav.c_uc=gps_gps_nav.c_uc;
  this->current_gps_nav.e=gps_gps_nav.e;
  this->current_gps_nav.c_us=gps_gps_nav.c_us;
  this->current_gps_nav.sqrt_a=gps_gps_nav.sqrt_a;
  this->current_gps_nav.t_oe=gps_gps_nav.t_oe;
  this->current_gps_nav.c_ic=gps_gps_nav.c_ic;
  this->current_gps_nav.omega_0=gps_gps_nav.omega_0;
  this->current_gps_nav.c_is=gps_gps_nav.c_is;
  this->current_gps_nav.i_0=gps_gps_nav.i_0;
  this->current_gps_nav.c_rc=gps_gps_nav.c_rc;
  this->current_gps_nav.omega=gps_gps_nav.omega;
  this->current_gps_nav.omega_dot=gps_gps_nav.omega_dot;
  this->current_gps_nav.i_dot=gps_gps_nav.i_dot;
  this->current_gps_nav.wnt_oc=gps_gps_nav.wnt_oc;
  this->current_gps_nav.wnt_oe=gps_gps_nav.wnt_oe;
  if(!this->event_server->event_is_set(this->new_gps_nav_block_event_id))
    this->event_server->set_event(new_gps_nav_block_event_id);
  this->nav_access.exit();
}

void CasteRx1::process_gps_alm(unsigned char *data,unsigned short int length)
{
  TGPSGPSAlm gps_gps_alm;

  this->nav_access.enter();
  memcpy((unsigned char *)&gps_gps_alm,data,sizeof(TGPSGPSAlm));
  this->current_gps_alm.gps_timestamp.wnc=gps_gps_alm.timestamp.wnc;
  this->current_gps_alm.gps_timestamp.tow=gps_gps_alm.timestamp.tow;
  this->current_gps_alm.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_gps_alm.sat_id=gps_gps_alm.prn;
  this->current_gps_alm.e=gps_gps_alm.e;
  this->current_gps_alm.t_oa=gps_gps_alm.t_oa;
  this->current_gps_alm.delta_i=gps_gps_alm.delta_i;
  this->current_gps_alm.omega_dot=gps_gps_alm.omega_dot;
  this->current_gps_alm.sqrt_a=gps_gps_alm.sqrt_a;
  this->current_gps_alm.omega_0=gps_gps_alm.omega_0;
  this->current_gps_alm.omega=gps_gps_alm.omega;
  this->current_gps_alm.m_0=gps_gps_alm.m_0;
  this->current_gps_alm.a_f1=gps_gps_alm.a_f1;
  this->current_gps_alm.a_f0=gps_gps_alm.a_f0;
  this->current_gps_alm.wn_a=gps_gps_alm.wn_a;
  this->current_gps_alm.as_config=gps_gps_alm.as_config;
  this->current_gps_alm.health8=gps_gps_alm.health8;
  this->current_gps_alm.health6=gps_gps_alm.health6;
  if(!this->event_server->event_is_set(this->new_gps_alm_block_event_id))
    this->event_server->set_event(new_gps_alm_block_event_id);
  this->nav_access.exit();
}

void CasteRx1::process_gps_ion(unsigned char *data,unsigned short int length)
{
  TGPSGPSIon gps_gps_ion;

  this->nav_access.enter();
  memcpy((unsigned char *)&gps_gps_ion,data,sizeof(TGPSGPSIon));
  this->current_gps_ion.gps_timestamp.wnc=gps_gps_ion.timestamp.wnc;
  this->current_gps_ion.gps_timestamp.tow=gps_gps_ion.timestamp.tow;
  this->current_gps_ion.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_gps_ion.sat_id=gps_gps_ion.prn;
  this->current_gps_ion.alpha_0=gps_gps_ion.alpha_0;
  this->current_gps_ion.alpha_1=gps_gps_ion.alpha_1;
  this->current_gps_ion.alpha_2=gps_gps_ion.alpha_2;
  this->current_gps_ion.alpha_3=gps_gps_ion.alpha_3;
  this->current_gps_ion.beta_0=gps_gps_ion.beta_0;
  this->current_gps_ion.beta_1=gps_gps_ion.beta_1;
  this->current_gps_ion.beta_2=gps_gps_ion.beta_2;
  this->current_gps_ion.beta_3=gps_gps_ion.beta_3;
  if(!this->event_server->event_is_set(this->new_gps_ion_block_event_id))
    this->event_server->set_event(new_gps_ion_block_event_id);
  this->nav_access.exit();
}

void CasteRx1::process_gps_utc(unsigned char *data,unsigned short int length)
{
  TGPSGPSUtc gps_gps_utc;

  this->nav_access.enter();
  memcpy((unsigned char *)&gps_gps_utc,data,sizeof(TGPSGPSUtc));
  this->current_gps_utc.gps_timestamp.wnc=gps_gps_utc.timestamp.wnc;
  this->current_gps_utc.gps_timestamp.tow=gps_gps_utc.timestamp.tow;
  this->current_gps_utc.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_gps_utc.sat_id=gps_gps_utc.prn;
  this->current_gps_utc.a_1=gps_gps_utc.a_1;
  this->current_gps_utc.a_0=gps_gps_utc.a_0;
  this->current_gps_utc.t_ot=gps_gps_utc.t_ot;
  this->current_gps_utc.wn_t=gps_gps_utc.wn_t;
  this->current_gps_utc.del_t_ls=gps_gps_utc.del_t_ls;
  this->current_gps_utc.wn_lsf=gps_gps_utc.wn_lsf;
  this->current_gps_utc.dn=gps_gps_utc.dn;
  this->current_gps_utc.del_t_lsf=gps_gps_utc.del_t_lsf;
  if(!this->event_server->event_is_set(this->new_gps_utc_block_event_id))
    this->event_server->set_event(new_gps_utc_block_event_id);
  this->nav_access.exit();
}

void CasteRx1::process_gps_raw_ca(unsigned char *data,unsigned short int length)
{
  TGPSRawCA gps_raw_ca;
  TGPSFrame new_frame;
  unsigned int i=0;

  this->nav_access.enter();
  memcpy((unsigned char *)&gps_raw_ca,data,sizeof(TGPSRawCA));
  new_frame.gps_timestamp.wnc=gps_raw_ca.timestamp.wnc;
  new_frame.gps_timestamp.tow=gps_raw_ca.timestamp.tow;
  new_frame.local_timestamp=this->local_time.getTimeInSeconds();
  new_frame.sat_id=gps_raw_ca.sat_id;
  new_frame.track_id=this->sat_channel[gps_raw_ca.sat_id];
  new_frame.wn=gps_raw_ca.timestamp.wnc;
  for(i=0;i<10;i++)
    new_frame.frame_bits[i]=gps_raw_ca.nav_bits[i];
  this->gps_raw_frames.push(new_frame);
  this->event_server->set_event(new_gps_raw_block_event_id);
  this->nav_access.exit();
}

void CasteRx1::process_meas_epoch(unsigned char *data,unsigned short int length)
{
  TGPSMeasEpoch gps_meas_epoch;
  TGPSMeasBlockType1 gps_meas_epoch_type1;
  TGPSMeasBlockType2 gps_meas_epoch_type2;
  unsigned short int index=0,i,j;
  TMeasEpochType1 meas_epoch_type1;
  TMeasEpochType2 meas_epoch_type2;
 
  this->meas_access.enter(); 
  memcpy((unsigned char *)&gps_meas_epoch,data,sizeof(TGPSMeasEpoch));
  index+=sizeof(TGPSMeasEpoch);
  this->current_meas_epoch.gps_timestamp.wnc=gps_meas_epoch.timestamp.wnc;
  this->current_meas_epoch.gps_timestamp.tow=gps_meas_epoch.timestamp.tow;
  this->current_meas_epoch.local_timestamp=this->local_time.getTimeInSeconds();
  if(gps_meas_epoch.common_flags&multipath_mitigation)
    this->current_meas_epoch.multipath_mitigation=true;
  else
    this->current_meas_epoch.multipath_mitigation=false;
  if(gps_meas_epoch.common_flags&smoothing_of_code)
    this->current_meas_epoch.smoothing_of_code=true;
  else
    this->current_meas_epoch.smoothing_of_code=false;
  this->current_meas_epoch.type1_data.clear();
  for(i=0;i<gps_meas_epoch.num_type1;i++)
  {
    memcpy((unsigned char *)&gps_meas_epoch_type1,&data[index],sizeof(TGPSMeasBlockType1));
    index+=gps_meas_epoch.length_type1; 
    meas_epoch_type1.rx_channel=gps_meas_epoch_type1.rx_channel;
    meas_epoch_type1.signal_type=(signal_t)gps_meas_epoch_type1.type.master_signal_type;
    meas_epoch_type1.antenna_id=(antenna_t)gps_meas_epoch_type1.type.antenna_id;
    meas_epoch_type1.sat_id=gps_meas_epoch_type1.svid;
    if(gps_meas_epoch_type1.code.msb==0 && gps_meas_epoch_type1.code.lsb==0)// the pseudo range is not valid
      meas_epoch_type1.pseudo_range=0.0;
    else
      meas_epoch_type1.pseudo_range=((double)gps_meas_epoch_type1.code.msb)*4294967.296+((double)gps_meas_epoch_type1.code.lsb)/1000.0;
    meas_epoch_type1.doppler=gps_meas_epoch_type1.doppler*0.0001;
    if(gps_meas_epoch_type1.carrier.msb==-128 && gps_meas_epoch_type1.carrier.lsb==0)// the carrier phase is not valid
      meas_epoch_type1.carrier_phase=0.0;
    else
      meas_epoch_type1.carrier_phase=(meas_epoch_type1.pseudo_range/299792458.0)+((double)gps_meas_epoch_type1.carrier.msb)*65.536+((double)gps_meas_epoch_type1.carrier.lsb)/1000.0;
    switch(meas_epoch_type1.signal_type)
    {
      case GPS_L1_CA:
      case GPS_L1_PY: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1575420000.0;
                      break;
      case GPS_L2_PY: 
      case GPS_L2C: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1227600000.0;
                    break;
      case GPS_L5: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1176450000.0;
                   break;
      case GLO_L1_CA: 
      case GLO_L1_P:
      case GLO_L2_P:
      case GLO_L2_CA: break;
      case GAL_L1A: 
      case GAL_L1BC: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1575420000.0;
                     break;
      case GAL_E6B:
      case GAL_E6BC: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1278750000.0;
                     break;
      case GAL_E5a: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1176450000.0;
                    break;
      case GAL_E5b: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1207140000.0;
                    break;
      case GAL_E5: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1191795000.0;
                   break;
      case GEO_L1CA: meas_epoch_type1.carrier_phase+=(meas_epoch_type1.pseudo_range/299792458.0)*1575420000.0;
                     break;
    }
    meas_epoch_type1.cn0=((double)gps_meas_epoch_type1.cn0)*0.25;
    if(meas_epoch_type1.signal_type==GPS_L1_PY || meas_epoch_type1.signal_type==GPS_L2_PY)
      meas_epoch_type1.cn0+=10;
    meas_epoch_type1.lock_time=gps_meas_epoch_type1.lock_time;
    if(gps_meas_epoch_type1.info.smoothed==1)
      meas_epoch_type1.pseudorange_smoothed=true;
    else
      meas_epoch_type1.pseudorange_smoothed=false;
    if(gps_meas_epoch_type1.info.smoothing_int_reached==1)
      meas_epoch_type1.smoothing_int_reached=true;
    else
      meas_epoch_type1.smoothing_int_reached=false;
    if(gps_meas_epoch_type1.info.phase_ambiguity==1)
      meas_epoch_type1.half_cycle_ambiguity=true;
    else
      meas_epoch_type1.half_cycle_ambiguity=false;
    meas_epoch_type1.type2_data.clear();
    for(j=0;j<gps_meas_epoch_type1.num_type2;j++)
    { 
      memcpy((unsigned char *)&gps_meas_epoch_type2,&data[index],sizeof(TGPSMeasBlockType2));
      index+=gps_meas_epoch.length_type2;
      meas_epoch_type2.signal_type=(signal_t)gps_meas_epoch_type2.type.master_signal_type;
      meas_epoch_type2.antenna_id=(antenna_t)gps_meas_epoch_type2.type.antenna_id;
      meas_epoch_type2.lock_time=gps_meas_epoch_type2.lock_time;
      meas_epoch_type2.cn0=((double)gps_meas_epoch_type2.cn0)*0.25;
      if(meas_epoch_type2.signal_type==GPS_L1_PY || meas_epoch_type2.signal_type==GPS_L2_PY)
        meas_epoch_type2.cn0+=10;
      if(gps_meas_epoch_type2.offsets_msb.code_offset_msb==-4 && gps_meas_epoch_type2.code_offset_lsb==0)// the pseudo range is not valid 
        meas_epoch_type2.pseudo_range=0.0;
      else
        meas_epoch_type2.pseudo_range=((double)gps_meas_epoch_type2.offsets_msb.code_offset_msb)*65.536+((double)gps_meas_epoch_type2.code_offset_lsb)/1000.0;
      if(gps_meas_epoch_type2.offsets_msb.doppler_offset_msb==-128 && gps_meas_epoch_type2.doppler_offset_lsb==0)// the doppler is not valid
        meas_epoch_type2.doppler=0.0;
      else
        meas_epoch_type2.doppler=((double)gps_meas_epoch_type2.offsets_msb.doppler_offset_msb)*6.5536+((double)gps_meas_epoch_type2.doppler_offset_lsb)/10000.0;
      if(gps_meas_epoch_type2.carrier_msb==-16 && gps_meas_epoch_type2.carrier_lsb==0)// the carrier phase is not valid
        meas_epoch_type2.carrier_phase=0.0;
      else
        meas_epoch_type2.carrier_phase=((double)gps_meas_epoch_type2.carrier_msb)*65.536+((double)gps_meas_epoch_type2.carrier_lsb)/1000.0;
      switch(meas_epoch_type2.signal_type)
      {
        case GPS_L1_CA:
        case GPS_L1_PY: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1575420000.0;
                        break;
        case GPS_L2_PY:
        case GPS_L2C: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1227600000.0;
                      break;
        case GPS_L5: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1176450000.0;
                     break;
        case GLO_L1_CA:
        case GLO_L1_P:
        case GLO_L2_P:
        case GLO_L2_CA: break;
        case GAL_L1A:
        case GAL_L1BC: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1575420000.0;
                       break;
        case GAL_E6B:
        case GAL_E6BC: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1278750000.0;
                       break;
        case GAL_E5a: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1176450000.0;
                      break;
        case GAL_E5b: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1207140000.0;
                      break;
        case GAL_E5: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1191795000.0;
                     break;
        case GEO_L1CA: meas_epoch_type2.carrier_phase+=(meas_epoch_type2.pseudo_range/299792458.0)*1575420000.0;
                       break;
      }
      if(gps_meas_epoch_type2.info.smoothed==1)
        meas_epoch_type2.pseudorange_smoothed=true;
      else
        meas_epoch_type2.pseudorange_smoothed=false;
      if(gps_meas_epoch_type2.info.smoothing_int_reached==1)
        meas_epoch_type2.smoothing_int_reached=true;
      else
        meas_epoch_type2.smoothing_int_reached=false;
      if(gps_meas_epoch_type2.info.phase_ambiguity==1)
        meas_epoch_type2.half_cycle_ambiguity=true;
      else
        meas_epoch_type2.half_cycle_ambiguity=false;
      meas_epoch_type1.type2_data.push_back(meas_epoch_type2);
    }
    this->current_meas_epoch.type1_data.push_back(meas_epoch_type1);
  }
  this->meas_access.exit();
}

void CasteRx1::process_meas_extra(unsigned char *data,unsigned short int length)
{
  TGPSMeasExtra gps_meas_extra;
  TGPSMeasExtraBlock gps_meas_extra_block;
  unsigned short int index=0,i;
  TMeasExtraBlock meas_extra_block;

  this->meas_access.enter();
  memcpy((unsigned char *)&gps_meas_extra,data,sizeof(TGPSMeasExtra));
  index+=sizeof(TGPSMeasExtra);
  this->current_meas_extra.gps_timestamp.wnc=gps_meas_extra.timestamp.wnc;
  this->current_meas_extra.gps_timestamp.tow=gps_meas_extra.timestamp.tow;
  this->current_meas_extra.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_meas_extra.doppler_var_factor=gps_meas_extra.doppler_var_factor;
  this->current_meas_extra.blocks.clear();
  for(i=0;i<gps_meas_extra.num_blocks;i++)
  {
    memcpy((unsigned char *)&gps_meas_extra_block,&data[index],sizeof(TGPSMeasExtraBlock));
    index+=gps_meas_extra.length_blocks;
    meas_extra_block.rx_channel=gps_meas_extra_block.rx_channel;
    meas_extra_block.signal_type=(signal_t)gps_meas_extra_block.type.master_signal_type;
    meas_extra_block.antenna_id=(antenna_t)gps_meas_extra_block.type.antenna_id;
    meas_extra_block.mp_correction=((double)gps_meas_extra_block.mp_correction)/1000.0;
    meas_extra_block.smoothing_correction=((double)gps_meas_extra_block.smoothing_correction)/1000.0;
    meas_extra_block.code_var=((double)gps_meas_extra_block.code_var)/10000.0;
    meas_extra_block.carrier_var=gps_meas_extra_block.carrier_var;
    meas_extra_block.lock_time=gps_meas_extra_block.lock_time;
    meas_extra_block.cum_lost_cont=gps_meas_extra_block.cum_lost_cont;
    this->current_meas_extra.blocks.push_back(meas_extra_block);
  }
  this->meas_access.exit();
}

void CasteRx1::process_end_of_meas(unsigned char *data,unsigned short int length)
{
  TGPSEndOfMeas gps_end_of_meas;

  memcpy((unsigned char *)&gps_end_of_meas,data,sizeof(TGPSEndOfMeas));
  if(!this->event_server->event_is_set(this->new_meas_block_event_id))
    this->event_server->set_event(new_meas_block_event_id);
}

void CasteRx1::process_pvt_cartesian(unsigned char *data,unsigned short int length)
{
  TGPSPVTCartesian gps_pvt_cartesian;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_cartesian,data,sizeof(TGPSPVTCartesian));
  this->current_pvt_cartesian.gps_timestamp.wnc=gps_pvt_cartesian.timestamp.wnc;
  this->current_pvt_cartesian.gps_timestamp.tow=gps_pvt_cartesian.timestamp.tow;
  this->current_pvt_cartesian.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_cartesian.mode=(pvt_mode_t)gps_pvt_cartesian.mode.mode;
  this->current_pvt_cartesian.solution=(solution_t)gps_pvt_cartesian.mode.solution_type;
  this->current_pvt_cartesian.error=gps_pvt_cartesian.error;
  this->current_pvt_cartesian.x=gps_pvt_cartesian.x;
  this->current_pvt_cartesian.y=gps_pvt_cartesian.y;
  this->current_pvt_cartesian.z=gps_pvt_cartesian.z;
  this->current_pvt_cartesian.undulation=gps_pvt_cartesian.undulation;
  this->current_pvt_cartesian.vx=gps_pvt_cartesian.vx;
  this->current_pvt_cartesian.vy=gps_pvt_cartesian.vy;
  this->current_pvt_cartesian.vz=gps_pvt_cartesian.vz;
  this->current_pvt_cartesian.cog=gps_pvt_cartesian.cog;
  this->current_pvt_cartesian.rx_clk_bias=gps_pvt_cartesian.rx_clk_bias;
  this->current_pvt_cartesian.rx_clk_drift=gps_pvt_cartesian.rx_clk_drift;
  this->current_pvt_cartesian.time_system=(sys_time_t)gps_pvt_cartesian.time_system;
  this->current_pvt_cartesian.datum=(datum_t)gps_pvt_cartesian.datum;
  this->current_pvt_cartesian.num_sat=gps_pvt_cartesian.num_sat;
  if(gps_pvt_cartesian.corr_info.orbit_clock_corr==1)
    this->current_pvt_cartesian.orbit_sat_clock_corr=true;
  else
    this->current_pvt_cartesian.orbit_sat_clock_corr=false;
  if(gps_pvt_cartesian.corr_info.range_corr==1)
    this->current_pvt_cartesian.range_corr_info=true;
  else
    this->current_pvt_cartesian.range_corr_info=false;
  if(gps_pvt_cartesian.corr_info.ion_info==1)
    this->current_pvt_cartesian.ion_info=true;
  else
    this->current_pvt_cartesian.ion_info=false;
  if(gps_pvt_cartesian.corr_info.orbit_acc_info==1)
    this->current_pvt_cartesian.orbit_accuracy_info=true;
  else
    this->current_pvt_cartesian.orbit_accuracy_info=false;
  if(gps_pvt_cartesian.corr_info.do229_mode==1)
    this->current_pvt_cartesian.DO299_used=true;
  else
    this->current_pvt_cartesian.DO299_used=false;
  this->current_pvt_cartesian.reference_id=gps_pvt_cartesian.reference_id;
  this->current_pvt_cartesian.mean_corr_age=gps_pvt_cartesian.mean_corr_age;
  this->current_pvt_cartesian.signal_info=gps_pvt_cartesian.signal_info;
  if(gps_pvt_cartesian.alert_flag.int_mon==1)
    this->current_pvt_cartesian.integrity_mon=true;
  else
    this->current_pvt_cartesian.integrity_mon=false;
  if(gps_pvt_cartesian.alert_flag.int_failed_raim==1)
    this->current_pvt_cartesian.integrity_failed_RAIM=true;
  else
    this->current_pvt_cartesian.integrity_failed_RAIM=false;
  if(gps_pvt_cartesian.alert_flag.int_failed_hpca==1)
    this->current_pvt_cartesian.integrity_failed_HPCA=true;
  else
    this->current_pvt_cartesian.integrity_failed_HPCA=false;
  this->pvt_access.exit();
}

void CasteRx1::process_pvt_geodetic(unsigned char *data,unsigned short int length)
{
  TGPSPVTGeodetic gps_pvt_geodetic;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_geodetic,data,sizeof(TGPSPVTGeodetic));
  this->current_pvt_geodetic.gps_timestamp.wnc=gps_pvt_geodetic.timestamp.wnc;
  this->current_pvt_geodetic.gps_timestamp.tow=gps_pvt_geodetic.timestamp.tow;
  this->current_pvt_geodetic.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_geodetic.mode=(pvt_mode_t)gps_pvt_geodetic.mode.mode;
  this->current_pvt_geodetic.solution=(solution_t)gps_pvt_geodetic.mode.solution_type;
  this->current_pvt_geodetic.error=gps_pvt_geodetic.error;
  this->current_pvt_geodetic.latitude=gps_pvt_geodetic.latitude;
  this->current_pvt_geodetic.longitude=gps_pvt_geodetic.longitude;
  this->current_pvt_geodetic.h=gps_pvt_geodetic.h;
  this->current_pvt_geodetic.undulation=gps_pvt_geodetic.undulation;
  this->current_pvt_geodetic.vn=gps_pvt_geodetic.vn;
  this->current_pvt_geodetic.ve=gps_pvt_geodetic.ve;
  this->current_pvt_geodetic.vu=gps_pvt_geodetic.vu;
  this->current_pvt_geodetic.cog=gps_pvt_geodetic.cog;
  this->current_pvt_geodetic.rx_clk_bias=gps_pvt_geodetic.rx_clk_bias;
  this->current_pvt_geodetic.rx_clk_drift=gps_pvt_geodetic.rx_clk_drift;
  this->current_pvt_geodetic.time_system=(sys_time_t)gps_pvt_geodetic.time_system;
  this->current_pvt_geodetic.datum=(datum_t)gps_pvt_geodetic.datum;
  this->current_pvt_geodetic.num_sat=gps_pvt_geodetic.num_sat;
  this->current_pvt_geodetic.wa_corr_info=gps_pvt_geodetic.wa_corr_info;
  if(gps_pvt_geodetic.corr_info.orbit_clock_corr==1)
    this->current_pvt_geodetic.orbit_sat_clock_corr=true;
  else
    this->current_pvt_geodetic.orbit_sat_clock_corr=false;
  if(gps_pvt_geodetic.corr_info.range_corr==1)
    this->current_pvt_geodetic.range_corr_info=true;
  else
    this->current_pvt_geodetic.range_corr_info=false;
  if(gps_pvt_geodetic.corr_info.ion_info==1)
    this->current_pvt_geodetic.ion_info=true;
  else
    this->current_pvt_geodetic.ion_info=false;
  if(gps_pvt_geodetic.corr_info.orbit_acc_info==1)
    this->current_pvt_geodetic.orbit_accuracy_info=true;
  else
    this->current_pvt_geodetic.orbit_accuracy_info=false;
  if(gps_pvt_geodetic.corr_info.do229_mode==1)
    this->current_pvt_geodetic.DO299_used=true;
  else
    this->current_pvt_geodetic.DO299_used=false;
  this->current_pvt_geodetic.reference_id=gps_pvt_geodetic.reference_id;
  this->current_pvt_geodetic.mean_corr_age=gps_pvt_geodetic.mean_corr_age;
  this->current_pvt_geodetic.signal_info=gps_pvt_geodetic.signal_info;
  if(gps_pvt_geodetic.alert_flag.int_mon==1)
    this->current_pvt_geodetic.integrity_mon=true;
  else
    this->current_pvt_geodetic.integrity_mon=false;
  if(gps_pvt_geodetic.alert_flag.int_failed_raim==1)
    this->current_pvt_geodetic.integrity_failed_RAIM=true;
  else
    this->current_pvt_geodetic.integrity_failed_RAIM=false;
  if(gps_pvt_geodetic.alert_flag.int_failed_hpca==1)
    this->current_pvt_geodetic.integrity_failed_HPCA=true;
  else
    this->current_pvt_geodetic.integrity_failed_HPCA=false;
  this->pvt_access.exit();
}

void CasteRx1::process_pos_cov_cartesian(unsigned char *data,unsigned short int length)
{
  TGPSPVTCov gps_pvt_pos_cov_cartesian;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_pos_cov_cartesian,data,sizeof(TGPSPVTCov));
  this->current_pvt_pos_cov_cartesian.gps_timestamp.wnc=gps_pvt_pos_cov_cartesian.timestamp.wnc;
  this->current_pvt_pos_cov_cartesian.gps_timestamp.tow=gps_pvt_pos_cov_cartesian.timestamp.tow;
  this->current_pvt_pos_cov_cartesian.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_pos_cov_cartesian.mode=(pvt_mode_t)gps_pvt_pos_cov_cartesian.mode.mode;
  this->current_pvt_pos_cov_cartesian.solution=(solution_t)gps_pvt_pos_cov_cartesian.mode.solution_type;
  this->current_pvt_pos_cov_cartesian.error=gps_pvt_pos_cov_cartesian.error;
  this->current_pvt_pos_cov_cartesian.cov_00=gps_pvt_pos_cov_cartesian.cov_00;
  this->current_pvt_pos_cov_cartesian.cov_11=gps_pvt_pos_cov_cartesian.cov_11;
  this->current_pvt_pos_cov_cartesian.cov_22=gps_pvt_pos_cov_cartesian.cov_22;
  this->current_pvt_pos_cov_cartesian.cov_33=gps_pvt_pos_cov_cartesian.cov_33;
  this->current_pvt_pos_cov_cartesian.cov_01=gps_pvt_pos_cov_cartesian.cov_01;
  this->current_pvt_pos_cov_cartesian.cov_02=gps_pvt_pos_cov_cartesian.cov_02;
  this->current_pvt_pos_cov_cartesian.cov_03=gps_pvt_pos_cov_cartesian.cov_03;
  this->current_pvt_pos_cov_cartesian.cov_12=gps_pvt_pos_cov_cartesian.cov_12;
  this->current_pvt_pos_cov_cartesian.cov_13=gps_pvt_pos_cov_cartesian.cov_13;
  this->current_pvt_pos_cov_cartesian.cov_23=gps_pvt_pos_cov_cartesian.cov_23;
  this->pvt_access.exit();
}

void CasteRx1::process_pos_cov_geodetic(unsigned char *data,unsigned short int length)
{
  TGPSPVTCov gps_pvt_pos_cov_geodetic;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_pos_cov_geodetic,data,sizeof(TGPSPVTCov));
  this->current_pvt_pos_cov_geodetic.gps_timestamp.wnc=gps_pvt_pos_cov_geodetic.timestamp.wnc;
  this->current_pvt_pos_cov_geodetic.gps_timestamp.tow=gps_pvt_pos_cov_geodetic.timestamp.tow;
  this->current_pvt_pos_cov_geodetic.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_pos_cov_geodetic.mode=(pvt_mode_t)gps_pvt_pos_cov_geodetic.mode.mode;
  this->current_pvt_pos_cov_geodetic.solution=(solution_t)gps_pvt_pos_cov_geodetic.mode.solution_type;
  this->current_pvt_pos_cov_geodetic.error=gps_pvt_pos_cov_geodetic.error;
  this->current_pvt_pos_cov_geodetic.cov_00=gps_pvt_pos_cov_geodetic.cov_00;
  this->current_pvt_pos_cov_geodetic.cov_11=gps_pvt_pos_cov_geodetic.cov_11;
  this->current_pvt_pos_cov_geodetic.cov_22=gps_pvt_pos_cov_geodetic.cov_22;
  this->current_pvt_pos_cov_geodetic.cov_33=gps_pvt_pos_cov_geodetic.cov_33;
  this->current_pvt_pos_cov_geodetic.cov_01=gps_pvt_pos_cov_geodetic.cov_01;
  this->current_pvt_pos_cov_geodetic.cov_02=gps_pvt_pos_cov_geodetic.cov_02;
  this->current_pvt_pos_cov_geodetic.cov_03=gps_pvt_pos_cov_geodetic.cov_03;
  this->current_pvt_pos_cov_geodetic.cov_12=gps_pvt_pos_cov_geodetic.cov_12;
  this->current_pvt_pos_cov_geodetic.cov_13=gps_pvt_pos_cov_geodetic.cov_13;
  this->current_pvt_pos_cov_geodetic.cov_23=gps_pvt_pos_cov_geodetic.cov_23;
  this->pvt_access.exit();
}

void CasteRx1::process_vel_cov_cartesian(unsigned char *data,unsigned short int length)
{
  TGPSPVTCov gps_pvt_vel_cov_cartesian;
  
  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_vel_cov_cartesian,data,sizeof(TGPSPVTCov));
  this->current_pvt_vel_cov_cartesian.gps_timestamp.wnc=gps_pvt_vel_cov_cartesian.timestamp.wnc;
  this->current_pvt_vel_cov_cartesian.gps_timestamp.tow=gps_pvt_vel_cov_cartesian.timestamp.tow;
  this->current_pvt_vel_cov_cartesian.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_vel_cov_cartesian.mode=(pvt_mode_t)gps_pvt_vel_cov_cartesian.mode.mode;
  this->current_pvt_vel_cov_cartesian.solution=(solution_t)gps_pvt_vel_cov_cartesian.mode.solution_type;
  this->current_pvt_vel_cov_cartesian.error=gps_pvt_vel_cov_cartesian.error;
  this->current_pvt_vel_cov_cartesian.cov_00=gps_pvt_vel_cov_cartesian.cov_00;
  this->current_pvt_vel_cov_cartesian.cov_11=gps_pvt_vel_cov_cartesian.cov_11;
  this->current_pvt_vel_cov_cartesian.cov_22=gps_pvt_vel_cov_cartesian.cov_22;
  this->current_pvt_vel_cov_cartesian.cov_33=gps_pvt_vel_cov_cartesian.cov_33;
  this->current_pvt_vel_cov_cartesian.cov_01=gps_pvt_vel_cov_cartesian.cov_01;
  this->current_pvt_vel_cov_cartesian.cov_02=gps_pvt_vel_cov_cartesian.cov_02;
  this->current_pvt_vel_cov_cartesian.cov_03=gps_pvt_vel_cov_cartesian.cov_03;
  this->current_pvt_vel_cov_cartesian.cov_12=gps_pvt_vel_cov_cartesian.cov_12;
  this->current_pvt_vel_cov_cartesian.cov_13=gps_pvt_vel_cov_cartesian.cov_13;
  this->current_pvt_vel_cov_cartesian.cov_23=gps_pvt_vel_cov_cartesian.cov_23;
  this->pvt_access.exit();
}

void CasteRx1::process_vel_cov_geodetic(unsigned char *data,unsigned short int length)
{
  TGPSPVTCov gps_pvt_vel_cov_geodetic;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_vel_cov_geodetic,data,sizeof(TGPSPVTCov));
  this->current_pvt_vel_cov_geodetic.gps_timestamp.wnc=gps_pvt_vel_cov_geodetic.timestamp.wnc;
  this->current_pvt_vel_cov_geodetic.gps_timestamp.tow=gps_pvt_vel_cov_geodetic.timestamp.tow;
  this->current_pvt_vel_cov_geodetic.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_vel_cov_geodetic.mode=(pvt_mode_t)gps_pvt_vel_cov_geodetic.mode.mode;
  this->current_pvt_vel_cov_geodetic.solution=(solution_t)gps_pvt_vel_cov_geodetic.mode.solution_type;
  this->current_pvt_vel_cov_geodetic.error=gps_pvt_vel_cov_geodetic.error;
  this->current_pvt_vel_cov_geodetic.cov_00=gps_pvt_vel_cov_geodetic.cov_00;
  this->current_pvt_vel_cov_geodetic.cov_11=gps_pvt_vel_cov_geodetic.cov_11;
  this->current_pvt_vel_cov_geodetic.cov_22=gps_pvt_vel_cov_geodetic.cov_22;
  this->current_pvt_vel_cov_geodetic.cov_33=gps_pvt_vel_cov_geodetic.cov_33;
  this->current_pvt_vel_cov_geodetic.cov_01=gps_pvt_vel_cov_geodetic.cov_01;
  this->current_pvt_vel_cov_geodetic.cov_02=gps_pvt_vel_cov_geodetic.cov_02;
  this->current_pvt_vel_cov_geodetic.cov_03=gps_pvt_vel_cov_geodetic.cov_03;
  this->current_pvt_vel_cov_geodetic.cov_12=gps_pvt_vel_cov_geodetic.cov_12;
  this->current_pvt_vel_cov_geodetic.cov_13=gps_pvt_vel_cov_geodetic.cov_13;
  this->current_pvt_vel_cov_geodetic.cov_23=gps_pvt_vel_cov_geodetic.cov_23;
  this->pvt_access.exit();
}

void CasteRx1::process_dop(unsigned char *data,unsigned short int length)
{
  TGPSPVTDOP gps_pvt_dop;

  this->pvt_access.enter();
  memcpy((unsigned char *)&gps_pvt_dop,data,sizeof(TGPSPVTDOP));
  this->current_pvt_dop.gps_timestamp.wnc=gps_pvt_dop.timestamp.wnc;
  this->current_pvt_dop.gps_timestamp.tow=gps_pvt_dop.timestamp.tow;
  this->current_pvt_dop.local_timestamp=this->local_time.getTimeInSeconds();
  this->current_pvt_dop.num_sat=gps_pvt_dop.num_sat;
  this->current_pvt_dop.p_dop=gps_pvt_dop.p_dop/100.0;
  this->current_pvt_dop.t_dop=gps_pvt_dop.t_dop/100.0;
  this->current_pvt_dop.t_dop=gps_pvt_dop.h_dop/100.0;
  this->current_pvt_dop.v_dop=gps_pvt_dop.v_dop/100.0;
  this->current_pvt_dop.hpl=gps_pvt_dop.hpl;
  this->current_pvt_dop.vpl=gps_pvt_dop.vpl;
  this->pvt_access.exit();
}

void CasteRx1::process_end_of_pvt(unsigned char *data,unsigned short int length)
{
  TGPSEndOfPVT gps_end_of_pvt;

  memcpy((unsigned char *)&gps_end_of_pvt,data,sizeof(TGPSEndOfMeas));
  if(!this->event_server->event_is_set(this->new_pvt_block_event_id))
    this->event_server->set_event(new_pvt_block_event_id);
}

unsigned short int CasteRx1::compute_crc(unsigned char *data,unsigned short int length)
{
  unsigned short int crc=0;
  unsigned int i=0;

  // compute the crc from part of the header
  for(i=0;i<4;i++)
    crc=(crc<<8)^CRC_16CCIT_LookUp[(crc>>8)^((unsigned char *)&this->current_header)[i+4]];
  // compute the crc from the data
  for(i=0;i<length;i++)
    crc=(crc<<8)^CRC_16CCIT_LookUp[(crc>>8)^data[i]];

  return crc;
}

void CasteRx1::sendCommand(const std::string &cmd)
{
  std::list<std::string> events;
  std::string reply;

  //relax if we concatenate commands. (200 ms were suggested by Septentrio support team in a personal communication). 50ms tested and it seems it is not enough!!!
  usleep(100000);
  
  events.push_back(this->new_reply_event_id);
  try{
    this->serialPort->write((unsigned char *)cmd.c_str(),cmd.size());
    this->event_server->wait_all(events,500);
    // compare the answer with the command
    reply=this->reply_queue.front();// get the reply
    this->reply_queue.pop();// remove it from the queue
    if(reply.find("Invalid command!")!=std::string::npos)
    {
      /* handle invalid command */
      throw CasteRx1Exception(_HERE_,"Invalid command");
    } 
    else if(reply.find(cmd.c_str(),0,cmd.size()-1)==std::string::npos)
    {
      /* handle unexpected answer */
      throw CasteRx1Exception(_HERE_,"Unexpected answer");
    }
  }catch(CEventTimeoutException &e){
    // no answer in the allowed time
    throw CasteRx1Exception(_HERE_,"Device did not answer in time");
  }
}

void CasteRx1::configureStreams(void)
{
  // stop any incomming data
  this->sendCommand("setSBFOutput,all,USB1,none\n");
  //enables septentrio binary format (SBF) as output at USB1
  this->sendCommand("setDataInOut,USB1,CMD,SBF\n");
  // create group 1 blocks (PVTCartesian,PosCovCartesian,DOP,PVTGeodetic)
  this->sendCommand("setSBFGroups,Group1,PVTCartesian+PosCovCartesian+VelCovCartesian+DOP+PVTGeodetic+PosCovGeodetic+VelCovGeodetic+EndOfPVT\n");
  // create group 2 blocks (GPSNav,GPSAlm,GPSIon,GPSUtc)
  #ifdef HAVE_GPSTK
  this->sendCommand("setSBFGroups,Group2,GPSRawCA\n");
  #else
  this->sendCommand("setSBFGroups,Group2,GPSNav+GPSAlm+GPSIon+GPSUtc\n");
  #endif
  // create group 3 blocks (MeasEpoch,MeasExtra,EndOfMeas)
  this->sendCommand("setSBFGroups,Group3,MeasEpoch+MeasExtra+EndOfMeas\n");
  // create group 4 blocks (ChannelStatus,ReceiverStatus,SatVisivility)
  this->sendCommand("setSBFGroups,Group4,ChannelStatus+ReceiverStatus+SatVisibility\n");
}

