#ifndef asterx1_gps_h
#define asterx1_gps_h

//include
#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <stdint.h>
#include "rs232.h"
#include "eventserver.h"
#include "threadserver.h"
#include "commexceptions.h"
#include "eventexceptions.h"
#include "gps_types.h"
#include "ctime.h"

typedef enum {header1,header2,get_cmd_reply,get_block_header,get_sbf_data} gps_state_t;

//physical constants
const double EARTH_RADIUS = 6378137.0;
const double EARTH_EXCENTRICITY = 0.0818191908426220; 

//Device protocol constants (SBF protocol)
const unsigned short int PVTCartesian_ID = 4006;
const unsigned short int PVTGeodetic_ID = 4007;
const unsigned short int PosCovCartesian_ID = 5905;
const unsigned short int PosCovGeodetic_ID = 5906;
const unsigned short int VelCovCartesian_ID = 5907;
const unsigned short int VelCovGeodetic_ID = 5908;
const unsigned short int DOP_ID = 4001;
const unsigned short int EndOfPVT_ID = 5921;
const unsigned short int MeasEpoch_ID = 4027;
const unsigned short int MeasExtra_ID = 4000;
const unsigned short int EndOfMeas_ID = 5922;
const unsigned short int GPSNav_ID = 5891;
const unsigned short int GPSAlm_ID = 5892;
const unsigned short int GPSIon_ID = 5893;
const unsigned short int GPSUtc_ID = 5894;
const unsigned short int GPSRawCA_ID = 4017;
const unsigned short int ReceiverStatus_ID = 4014;
const unsigned short int ChannelStatus_ID = 4013;
const unsigned short int SatVisibility_ID = 4012;

//acquisition constants
const int DATA_TIMEOUT = 11000; //msec
const int CONFIG_TIMEOUT = 500; //msec

//data sizes
const unsigned int BLOCK_HEADER_SIZE = 8;

/**
  CasteRx1 implements continuous data acquisition from the asteRx1 GPS receiver. 
  Serial communications are implemented by the class CRS232, inherited from CComm (both belonging to IRI communication library)
  Detailed documentation about SBF data transfer protocol can be found in the device manual.
  PLEASE NOTE: This driver only implements a part of the AsteRx1 device functionalities (the most useful part for mobile robot navigation).
  PLEASE NOTE: For extended functionalities , please refer to the manual.
Programmer: Andreu Corominas Murtra, acorominas@iri.upc.edu
Device website: www.septentrio.com
 */
class CasteRx1
{
  public:
    /** \brief Default constructor
     *
     * Constructor. Requires port file name and, optionally, acquisition rate.
     * 
     */
    CasteRx1(const std::string &name,const std::string &hwPortName = "/dev/ttyACM0");
    /** \brief Destructor
     * 
     * Destructor: Stops acquisition in case it was running and closes serial Port in case it was opened.
     * 
     */
    virtual ~CasteRx1(); 
    /** \brief Open and configure serial device. 
     * 
     * Open and configure serial device to 8N1 Byte transmission. 
     * Launches a thread for serial in/out comms
     */
    void openDevice(); 
    /** \brief Closes serial comm
     * 
     * Closes serial comm's , stops comm thread and deletes comm object.
     * If success , resets status bit DEVICE_OPEN.
     */
    void closeDevice();
    /** \brief Starts basic data acquisition
     * 
     * Configures the device to perform some signal processing basics for mobile robotics.
     * Starts a mode of operation of continuos data flow from the device at a rate of 2Hz.
     */
    void startAcquisition(); 
    /** \brief Stops acquisition
     * 
     * Stops continuous acquisition mode
     */ 
    void stopAcquisition(); 
    /**
     * \brief
     *
     */
    std::string get_new_meas_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_meas_data(TMeasEpoch &meas_epoch, TMeasExtra &meas_extra);
    /**
     * \brief
     *
     */
    std::string get_new_gps_nav_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_gps_nav_data(TGPSNav &gps_nav);
    /**
     * \brief
     *
     */
    std::string get_new_gps_alm_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_gps_alm_data(TGPSAlm &gps_alm);
    /**
     * \brief
     *
     */
    std::string get_new_gps_ion_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_gps_ion_data(TGPSIon &gps_ion);
    /**
     * \brief
     *
     */
    std::string get_new_gps_utc_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_gps_utc_data(TGPSUtc &gps_utc);
    /**
     * \brief
     *
     */
    std::string get_new_gps_raw_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_gps_raw_data(TGPSFrame &gps_frame);
    /**
     * \brief
     *
     */
    std::string get_new_pvt_data_event_id(void);
    /**
     * \brief
     *
     */
    void get_pvt_data(TPVTCartesian &pvt_cart,TPVTGeodetic &pvt_geo);
    /**
     * \brief
     *
     */
    void get_pvt_cov_data(TPVTCov &pvt_pos_cov_cart,TPVTCov &pvt_pos_cov_geo,TPVTCov &pvt_vel_cov_cart,TPVTCov &pvt_vel_cov_geo);
    /**
     * \brief
     *
     */
    void get_pvt_dop_data(TPVTDOP &pvt_dop);
  private:
    CRS232 *serialPort; /**<Object for serial comms. This object launch an independent thread managing byte I/O through serial port (USB or RS232)*/
    std::string serial_device;
    /* internal queue for command replies */
    std::queue<std::string> reply_queue;
    /* event server and events */
    CEventServer *event_server;
    std::string finish_thread_event_id;
    std::string new_reply_event_id;
    /* thread server and threads */
    CThreadServer *thread_server;
    std::string gps_process_data_thread_id;
    bool gps_running;
    CTime local_time;
    std::vector<unsigned short int> sat_channel;

    /* header of the message currently being received */
    TGPSBlockHeader current_header;
    TReceiverStatus receiver_status;
    TSatInfo sat_info;
 
    /* Meas blocks attributes */
    CMutex meas_access;
    std::string new_meas_block_event_id;
    TMeasEpoch current_meas_epoch;
    TMeasExtra current_meas_extra;

    /* GPS naviagtion blocks attributes */
    CMutex nav_access;
    std::string new_gps_nav_block_event_id;
    TGPSNav current_gps_nav;
    std::string new_gps_alm_block_event_id;
    TGPSAlm current_gps_alm;
    std::string new_gps_ion_block_event_id;
    TGPSIon current_gps_ion;
    std::string new_gps_utc_block_event_id;
    TGPSUtc current_gps_utc;
    std::string new_gps_raw_block_event_id;
    std::queue<TGPSFrame> gps_raw_frames;

    /* PVT blocks attributes */
    CMutex pvt_access;
    std::string new_pvt_block_event_id;
    TPVTCartesian current_pvt_cartesian;
    TPVTGeodetic current_pvt_geodetic;
    TPVTCov current_pvt_pos_cov_cartesian;
    TPVTCov current_pvt_vel_cov_cartesian;
    TPVTCov current_pvt_pos_cov_geodetic;
    TPVTCov current_pvt_vel_cov_geodetic;
    TPVTDOP current_pvt_dop;
  protected:
    /**
     * \brief
     *
     */
    static void *gps_process_data_thread(void *param);
    /**
     * \brief
     *
     */
    void process_block(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_channel_status(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_receiver_status(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_sat_visibility(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_gps_nav(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_gps_alm(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_gps_ion(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_gps_utc(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_gps_raw_ca(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_meas_epoch(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_meas_extra(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_end_of_meas(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_pvt_cartesian(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_pos_cov_cartesian(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_pos_cov_geodetic(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_vel_cov_cartesian(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_vel_cov_geodetic(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_pvt_geodetic(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_dop(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void process_end_of_pvt(unsigned char *data,unsigned short int length);
    /**
     * \brief
     *
     */
    void configureStreams(void);
    /**
     * \brief
     *
     */
    void sendCommand(const std::string &cmnd); 

    /**
      computes the crc-ccitt of buf with polynomic 0x1021
     */
    unsigned short int compute_crc(unsigned char *data,unsigned short int length);
};

#endif
