#ifndef _GPS_TYPES_H
#define _GPS_TYPES_H

// global satellite information
typedef struct{
  unsigned int tow;
  unsigned short int wnc;
}TBlockTimeStamp;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  long int frame_bits[10];
  int wn;
  short int sat_id;
  short int track_id;
}TGPSFrame;

// User data structures
typedef enum {elev_setting=0,elev_rising=1,elev_unknown=3} rise_set_t;

typedef enum {health_unknown=0, healthy=1, unhealthy=3} health_t;

typedef enum {track_idle=0, track_search=1, track_sync=2, tracking=3} track_status_t;

typedef enum {pvt_not_used=0, pvt_waiting=1,pvt_used=2, pvt_rejected=3} pvt_status_t;

typedef enum {main_antenna=0,aux1_antenna=1,aux2_antenna=2} antenna_t;

typedef struct {
  antenna_t antenna_id;
  track_status_t track_status;
  pvt_status_t pvt_status;
}TChannelStateInfo;

typedef struct {
  int sat_id;
  unsigned int azimuth;
  rise_set_t rise_set;
  health_t health;
  int elevation;
  int rx_channel;
  std::vector<TChannelStateInfo> state;
}TChannelSatInfo;

typedef struct {
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  std::vector<TChannelSatInfo> info;
}TChannelStatus;

typedef enum {gps_sbas_galileo_l1=0,glonass_l1=1,galileo_e6=2,gps_l2=3,glonass_l2=4,gps_sbas_galileo_l5_e5a=5,galileo_e5b=6,galileo_e5a_b=7} frontend_t;

enum {vbat_ok=0x00000001,
      antenna_connected=0x00000001,
      ext_ref=0x00000002,
      pps_in=0x00000004,
      wn_set=0x00000008,
      tow_set=0x00000010,
      fine_time=0x00000020,
      disk_activity=0x00000040,
      disk_full=0x00000080,
      disk_mounted=0x00000100};

enum {nvram=0x00000001,
      baseband=0x00000002,
      rtc=0x00000004,
      software=0x00000008,
      watchdog=0x00000010,
      antenna_ok=0x00000020,
      congestion=0x00000040,
      sis_error=0x00000080,
      missed_event=0x00000100,
      cpu_overload=0x00000200,
      invalid_config=0x00000400};

typedef struct {
  frontend_t frontend_id;
  antenna_t antenna_id;
  int gain;
  int sample_var;
  int blanking_stat;
}TAGCData;

typedef struct {
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int cpu_load;
  unsigned int up_time;
  unsigned int state;
  unsigned int error;
  std::vector<TAGCData> agc;
}TReceiverStatus;

typedef enum {almanac=1, ephemeris=2, unkown_info=255} sat_info_t;

typedef struct {
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int sat_id;
  double azimuth;
  double elevation;
  rise_set_t rise_set;
  sat_info_t sat_info;
}TSatInfo;

typedef enum {GPS_L1_CA=0,GPS_L1_PY=1,GPS_L2_PY=2,GPS_L2C=3,GPS_L5=4,GLO_L1_CA=8,GLO_L1_P=9,GLO_L2_P=10,GLO_L2_CA=11,GAL_L1A=16,
              GAL_L1BC=17,GAL_E6B=18,GAL_E6BC=19,GAL_E5a=20,GAL_E5b=21,GAL_E5=22,GEO_L1CA=24} signal_t;

const std::string signal_type_name[] = {"GPS_L1_CA (1575.42 MHz)",
                                        "GPS_L1_P(Y) (1575.42 MHz)",
                                        "GPS_L2_P(Y) (1227.60 MHz)",
                                        "GPS_L2C (1227.60 MHz)",
                                        "GPS_L5 (1176.45 MHz)",
                                        "",
                                        "",
                                        "",
                                        "GLO_L1_CA",
                                        "GLO_L1_P",
                                        "GLO_L2_P",
                                        "GLO_L1_CA",
                                        "",
                                        "",
                                        "",
                                        "",
                                        "GAL_L1A (1575.42 MHz)",
                                        "GAL_L1BC (1575.42 MHz)",
                                        "GAL_E6A (1278.75 MHz)",
                                        "GAL_E6BC (1278.75 MHz)",
                                        "GAL_E5a (1176.45 MHz)",
                                        "GAL_E5b (1207.14 MHz)",
                                        "GAL_E5 (1191.795 MHz)",
                                        "",
                                        "GEO_L1CA (1575.42 MHz)"};
 
typedef struct{
  signal_t signal_type;
  antenna_t antenna_id;
  unsigned int lock_time;
  double cn0;
  double pseudo_range;
  double doppler;
  double carrier_phase;
  bool pseudorange_smoothed;
  bool smoothing_int_reached;
  bool half_cycle_ambiguity;
}TMeasEpochType2;

typedef struct{
  unsigned int rx_channel;
  signal_t signal_type;
  antenna_t antenna_id;
  unsigned int sat_id;
  double pseudo_range;
  double doppler;
  double carrier_phase;
  double cn0;
  unsigned int lock_time;
  bool pseudorange_smoothed;
  bool smoothing_int_reached;
  bool half_cycle_ambiguity;
  std::vector<TMeasEpochType2> type2_data;
}TMeasEpochType1;

enum {multipath_mitigation=0x01,smoothing_of_code=0x02};

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  bool multipath_mitigation;
  bool smoothing_of_code;
  std::vector<TMeasEpochType1> type1_data;
}TMeasEpoch;

typedef struct{
  unsigned int rx_channel;
  signal_t signal_type;
  antenna_t antenna_id;
  int mp_correction;
  int smoothing_correction;
  unsigned int code_var;
  unsigned int carrier_var;
  unsigned int lock_time;
  unsigned int cum_lost_cont;
}TMeasExtraBlock;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  float doppler_var_factor;
  std::vector<TMeasExtraBlock> blocks;
}TMeasExtra;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int sat_id;
  int wn;
  unsigned int ca_or_pon_l2;
  unsigned int ura;
  unsigned int health;
  unsigned int l2_data_flag;
  unsigned int iodc;
  unsigned int iode2;
  unsigned int iode3;
  unsigned int fit_int_flag;
  double t_gd;
  unsigned int t_oc;
  double a_f2;
  double a_f1;
  double a_f0;
  double c_rs;
  double delta_n;
  double m_0;
  double c_uc;
  double e;
  double c_us;
  double sqrt_a;
  unsigned int t_oe;
  double c_ic;
  double omega_0;
  double c_is;  
  double i_0;
  double c_rc;
  double omega;
  double omega_dot;
  double i_dot;
  unsigned int wnt_oc;
  unsigned int wnt_oe;
}TGPSNav;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int sat_id;
  double e;
  unsigned int t_oa;
  double delta_i;
  double omega_dot;
  double sqrt_a;
  double omega_0;
  double omega;
  double m_0;
  double a_f1;
  double a_f0;
  unsigned int wn_a;
  unsigned int as_config;
  unsigned int health8;
  unsigned int health6;
}TGPSAlm;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int sat_id;
  double alpha_0;
  double alpha_1;
  double alpha_2;
  double alpha_3;
  double beta_0;
  double beta_1;
  double beta_2;
  double beta_3;
}TGPSIon;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int sat_id;
  double a_1;
  double a_0;
  unsigned int t_ot;
  unsigned int wn_t;
  int del_t_ls;
  unsigned int wn_lsf;
  unsigned int dn;
  int del_t_lsf;
}TGPSUtc;

typedef enum {no_pvt=0,stand_alone_pvt=1,diff_pvt=2,fixed_loc=3,rtk_fixed_amb=4,rtk_float_amb=5,sbas_pvt=6} pvt_mode_t;

typedef enum {pvt_3d=0,pvt_2d=1} solution_t;

enum {no_error=0,not_enough_meas=1,not_enough_eph=2,dop_too_large=3,res_too_large=4,no_conv=5,not_enough_meas_outliers=6,export_laws=7,not_enough_diff_corr=8,no_base_coord=9};

typedef enum {gps_time=0,galileo_time=1} sys_time_t;

typedef enum {wgs_84=0,galileo=1,pz_90_02=2,user_datum1=250,user_datum2=251} datum_t;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  pvt_mode_t mode;
  solution_t solution;
  unsigned int error;
  double x;
  double y;
  double z;
  double undulation;
  double vx;
  double vy;
  double vz;
  double cog;
  double rx_clk_bias;
  double rx_clk_drift;
  sys_time_t time_system;
  datum_t datum;
  unsigned int num_sat;
  bool orbit_sat_clock_corr;
  bool range_corr_info;
  bool ion_info;
  bool orbit_accuracy_info;
  bool DO299_used;
  unsigned int reference_id;
  unsigned int mean_corr_age;
  unsigned int signal_info;
  bool integrity_mon;
  bool integrity_failed_RAIM;
  bool integrity_failed_HPCA;
}TPVTCartesian;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  pvt_mode_t mode;
  solution_t solution;
  unsigned int error;
  double latitude;
  double longitude;
  double h;
  double undulation;
  double vn;
  double ve;
  double vu;
  double cog;
  double rx_clk_bias;
  double rx_clk_drift;
  sys_time_t time_system;
  datum_t datum;
  unsigned int num_sat;
  unsigned int wa_corr_info;
  bool orbit_sat_clock_corr;
  bool range_corr_info;
  bool ion_info;
  bool orbit_accuracy_info;
  bool DO299_used;
  unsigned int reference_id;
  unsigned int mean_corr_age;
  unsigned int signal_info;
  bool integrity_mon;
  bool integrity_failed_RAIM;
  bool integrity_failed_HPCA;
}TPVTGeodetic;

typedef struct{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  pvt_mode_t mode;
  solution_t solution;
  unsigned int error;
  double cov_00;
  double cov_11;
  double cov_22;
  double cov_33;
  double cov_01;
  double cov_02;
  double cov_03;
  double cov_12;
  double cov_13;
  double cov_23;
}TPVTCov;

typedef struct
{
  TBlockTimeStamp gps_timestamp;
  double local_timestamp;
  unsigned int num_sat;
  double p_dop;
  double t_dop;
  double h_dop;
  double v_dop;
  double hpl;
  double vpl;
}TPVTDOP;

// these precompiler commands force the byte aligment in memory of the data structures
//#pragma pack (push, 1)
//#pragma pack (pop)

// block ID structure
#pragma pack (push, 1)
typedef struct{
  uint16_t id:13;
  uint16_t rev:3;
}TGPSBlockId;
#pragma pack (pop)

// header structure
#pragma pack (push, 1)
typedef struct{
  uint8_t sync[2];
  uint16_t crc;
  TGPSBlockId id;
  uint16_t length;
}TGPSBlockHeader;
#pragma pack (pop)

// block time stamp structure
#pragma pack (push, 1)
typedef struct{
  uint32_t tow;
  uint16_t wnc;
}TGPSBlockTimeStamp;
#pragma pack (pop)

// Measure Epoch structure
#pragma pack (push, 1)
typedef struct{
  uint8_t master_signal_type:5;
  uint8_t antenna_id:3;
}TGPSMeasBlockType;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t smoothed:1;
  uint8_t smoothing_int_reached:1;
  uint8_t phase_ambiguity:1;
  uint8_t freq_nr:5;
}TGPSMeasBlockObsInfo;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t code_offset_msb:3;
  uint8_t doppler_offset_msb:5;
}TGPSMeasBlockType2Offsets;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSMeasBlockType type;
  uint8_t lock_time;
  uint8_t cn0;
  TGPSMeasBlockType2Offsets offsets_msb;
  int8_t carrier_msb;
  TGPSMeasBlockObsInfo info;
  uint16_t code_offset_lsb;
  uint16_t carrier_lsb;
  uint16_t doppler_offset_lsb; 
}TGPSMeasBlockType2;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t msb:4;
  uint8_t reserved:4;
  uint32_t lsb;
}TGPSMeasBlockType1Code;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint16_t lsb;
  int8_t msb;
}TGPSMeasBlockType1Carrier;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t rx_channel;
  TGPSMeasBlockType type;
  uint8_t svid;
  TGPSMeasBlockType1Code code;
  int32_t doppler;
  TGPSMeasBlockType1Carrier carrier;
  uint8_t cn0;
  uint16_t lock_time;
  TGPSMeasBlockObsInfo info;
  uint8_t num_type2;
}TGPSMeasBlockType1;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t num_type1;
  uint8_t length_type1;
  uint8_t length_type2;
  uint8_t common_flags;
  uint8_t reserved[2];
}TGPSMeasEpoch;
#pragma pack (pop)

// Measurement extra structure
#pragma pack (push, 1)
typedef struct{
  uint8_t rx_channel;
  TGPSMeasBlockType type;
  int16_t mp_correction;
  int16_t smoothing_correction;
  uint16_t code_var;
  uint16_t carrier_var;
  uint16_t lock_time;
  uint8_t cum_lost_cont;
  uint8_t reserved[3];
}TGPSMeasExtraBlock;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t num_blocks;
  uint8_t length_blocks;
  float doppler_var_factor;
}TGPSMeasExtra;
#pragma pack (pop)

// measurement end structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
}TGPSEndOfMeas;
#pragma pack (pop)

// GPS NAV structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t prn;
  uint8_t reserved1;
  int16_t wn;
  uint8_t ca_or_pon_l2;
  uint8_t ura;
  uint8_t health;
  uint8_t l2_data_flag;
  uint16_t iodc;
  uint8_t iode2;
  uint8_t iode3;
  uint8_t fit_int_flag;
  uint8_t reserved2;
  float t_gd;
  uint32_t t_oc;
  float a_f2;
  float a_f1;
  float a_f0;
  float c_rs;
  float delta_n;
  double m_0;
  float c_uc;
  double e;
  float c_us;
  double sqrt_a;
  uint32_t t_oe;
  float c_ic;
  double omega_0;
  float c_is;  
  double i_0;
  float c_rc;
  double omega;
  float omega_dot;
  float i_dot;
  uint16_t wnt_oc;
  uint16_t wnt_oe;
}TGPSGPSNav;
#pragma pack (pop)

// APG Alm structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t prn;
  uint8_t reserved1;
  float e;
  uint32_t t_oa;
  float delta_i;
  float omega_dot;
  float sqrt_a;
  float omega_0;
  float omega;
  float m_0;
  float a_f1;
  float a_f0;
  uint8_t wn_a;
  uint8_t as_config;
  uint8_t health8;
  uint8_t health6;
}TGPSGPSAlm;
#pragma pack (pop)

// GPS Ion structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t prn;
  uint8_t reserved1;
  float alpha_0;
  float alpha_1;
  float alpha_2;
  float alpha_3;
  float beta_0;
  float beta_1;
  float beta_2;
  float beta_3;
}TGPSGPSIon;
#pragma pack (pop)

// GPs UTC structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t prn;
  uint8_t reserved1;
  float a_1;
  double a_0;
  uint32_t t_ot;
  uint8_t wn_t;
  int8_t del_t_ls;
  uint8_t wn_lsf;
  uint8_t dn;
  int8_t del_t_lsf;
}TGPSGPSUtc;
#pragma pack (pop)

// Position, velocity and time Cartesian structure
#pragma pack (push, 1)
typedef struct{
  uint8_t mode:4;
  uint8_t reserved:3;
  uint8_t solution_type:1;
}TGPSPVTMode;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t orbit_clock_corr:1;
  uint8_t range_corr:1;
  uint8_t ion_info:1;
  uint8_t orbit_acc_info:1;
  uint8_t do229_mode:1;
  uint8_t reserver:3;
}TGPSPVTInfo;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t int_mon:1;
  uint8_t int_failed_raim:1;
  uint8_t int_failed_hpca:1;
  uint8_t galileo_storm_flag:1;
  uint8_t reserved:4;
}TGPSPVTAlert;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  TGPSPVTMode mode;
  uint8_t error;
  double x;
  double y;
  double z;
  float undulation;
  float vx;
  float vy;
  float vz;
  float cog;
  double rx_clk_bias;
  float rx_clk_drift;
  uint8_t time_system;
  uint8_t datum;
  uint8_t num_sat;
  uint8_t wa_corr_info;
  TGPSPVTInfo corr_info;
  uint16_t reference_id;
  uint16_t mean_corr_age;
  uint32_t signal_info;
  TGPSPVTAlert alert_flag;
  uint8_t reserved[3];
}TGPSPVTCartesian;
#pragma pack (pop)

// Position, velocity and time Geodetic structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  TGPSPVTMode mode;
  uint8_t error;
  double latitude;
  double longitude;
  double h;
  float undulation;
  float vn;
  float ve;
  float vu;
  float cog;
  double rx_clk_bias;
  float rx_clk_drift;
  uint8_t time_system;
  uint8_t datum;
  uint8_t num_sat;
  uint8_t wa_corr_info;
  TGPSPVTInfo corr_info;
  uint16_t reference_id;
  uint16_t mean_corr_age;
  uint32_t signal_info;
  TGPSPVTAlert alert_flag;
  uint8_t reserved[3];
}TGPSPVTGeodetic;
#pragma pack (pop)

// Position, velocity and time Covariance Cartesian structure
#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  TGPSPVTMode mode;
  uint8_t error;
  float cov_00;
  float cov_11;
  float cov_22;
  float cov_33;
  float cov_01;
  float cov_02;
  float cov_03;
  float cov_12;
  float cov_13;
  float cov_23;
}TGPSPVTCov;
#pragma pack (pop)

// Position, velocity and time DOP structure
#pragma pack (push, 1)
typedef struct
{
  TGPSBlockTimeStamp timestamp;
  uint8_t num_sat;
  uint8_t reserved;
  uint16_t p_dop;
  uint16_t t_dop;
  uint16_t h_dop;
  uint16_t v_dop;
  float hpl;
  float vpl; 
}TGPSPVTDOP;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
}TGPSEndOfPVT;
#pragma pack (pop)

// channel status structure
#pragma pack (push, 1)
typedef struct{
  uint8_t antenna;
  uint8_t reserved;
  uint16_t tracking_status;
  uint16_t pvt_status;
  uint16_t pvy_info;
}TGPSChannelStateInfo;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint16_t azimuth:9;
  uint16_t reserved:5;
  uint16_t rise_set:2;
}TGPSAzimuthRiseSet;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t sat_id;
  uint8_t freq_nr;
  uint16_t reserved1;
  TGPSAzimuthRiseSet azimuth_rise_set;
  uint16_t health;
  int8_t elevation;
  uint8_t num_state_info;
  uint8_t rx_channel;
  uint8_t reserved2;
}TGPSChannelSatInfo;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t num_ch;
  uint8_t length_ch_sat_info;
  uint8_t length_ch_state_info;
  uint8_t reserved[3];
}TGPSChannelStatus;
#pragma pack (pop)

// Receiver status structure
#pragma pack (push, 1)
typedef struct{
  uint32_t vbat_ok:1;
  uint32_t antenna_connected:1;
  uint32_t ext_ref:1;
  uint32_t pps_in:1;
  uint32_t wn_set:1;
  uint32_t tow_set:1;
  uint32_t fine_time:1;
  uint32_t disk_activity:1;
  uint32_t disk_full:1;
  uint32_t disk_mounted:1;
  uint32_t reserved:22;
}TGPSReceiverStatusRxState;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint32_t nvram:1;
  uint32_t baseband:1;
  uint32_t rtc:1;
  uint32_t software:1;
  uint32_t watchdog:1;
  uint32_t antenna_ok:1;
  uint32_t congestion:1;
  uint32_t sis_error:1; 
  uint32_t missed_event:1;
  uint32_t cpu_overload:1;
  uint32_t invalid_config:1;
  uint32_t reserved:21;
}TGPSReceiverStatusRxError;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  uint8_t code:4;
  uint8_t antenna_id:4;
}TGPSAGCDataFrontendId;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSAGCDataFrontendId frontend_id;
  int8_t gain;
  uint8_t sample_var;
  uint8_t blanking_stat;
}TGPSAGCData;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t cpu_load;
  uint8_t reserved1;
  uint32_t up_time;
  TGPSReceiverStatusRxState rx_state;
  TGPSReceiverStatusRxError rx_error;
  uint8_t num_agc_states;
  uint8_t length_agc_states;
  uint8_t reserved2[2];
}TGPSReceiverStatus;
#pragma pack (pop)

// satellite visibility structure
#pragma pack (push, 1)
typedef struct{
  uint8_t sat_id;
  uint8_t freq_nr;
  uint16_t azimuth;
  int16_t elevation;
  uint8_t rise_set;
  uint8_t sat_info;
}TGPSSatInfo;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t num_sat;
  uint8_t length_sat;
}TGPSSatVisibility;
#pragma pack (pop)

#pragma pack (push, 1)
typedef struct{
  TGPSBlockTimeStamp timestamp;
  uint8_t sat_id;
  uint8_t crc_passed;
  uint8_t viterbi_count;
  uint8_t source;
  uint8_t freq_nr;
  uint8_t reserved;
  uint32_t nav_bits[10];
}TGPSRawCA;
#pragma pack (pop)

#endif
